#!/usr/bin/python
'''Usage: mythautoicon.py [OPTIONS]
MythAutoIcon v%s
For fuzzy matching channels in MythTV database and grabbing icons for them.
Options:
    -h, --help              Print this info.
    -r, --regions           Comma seperated region list, comprising type:region,
                            i.e. -r lsc:us-il,lsc:us-ma,lsc:us,lss:astra1c
                            (Default: lsc:uk)
    -s, --source=SOURCE     Only grab icons for channels provided by SOURCE ID.
                            (Default: All sources)
    -o, --overwrite         Create all new data for regions. (This will destroy
                            old settings!)
    -c, --callsigns         Work in callsign mode. (Do not d/l icons, let
                            mythfilldatabase do it using iconmap tables)
    -f, --force-update      Override the built in caching of the web source.
    -n, --no-auto-select    Auto selecting exact matches is turned off.
    -d, --dump=FILE         Don't perform SQL changes, dump them to FILE
                            instead.
    -i, --info              Show channels against index from DB query.
                            Exclusive, exits without further processing.
    -l, --list=LIST         Specify list of channel index's (see -i output) to
                            affect i.e. 4,7,8-14,20-27.
    -e, --erase             Remove icon paths from database. Helps for forcing
                            re-download of icons.
    -u, --unassigned        Only process channels with no existing value for
                            icon in the database.

Steps:
    1. Grab list of stations from DB.
    2. Grab list of icons available from web source.
    3. For each station, do a fuzzy match.
    If in callsign mode:
        4. Update the tables in DB.
           (You then run "mythfilldatabase --update-icon-map".)
    else:
        4. Download icons.
        5. Insert full path to icon into DB.
           (You then move the dowloaded icons to the appropriate directory.)
    6. Update the cache and save to disk.
    
Note: Lots of places don't use callsigns, hence they are optional. As they
      aren't used here in the UK, they are also EXPERIMENTAL. YMMV etc.
'''
__revision__ = '0.2'
# Coming soon.... maybe...
#   -e, --equiv       Treat channels with same name as equivalent, and fill with
#                           same icon if already selected. i.e. PPV

# # # # # # User Settings Here # # # # # #
# This is probably the main one you'll want to change so that you don't have to keep specifying it
default_region_codes = 'lsc:uk'
# path that icons will be located in on mythtv system (manually copy icons to this path, this is
# what is placed in the DB)
icon_dir_mythtv = '/mnt/store/mythtv/.mythtv/icons'
# icons are downloaded here (might not be running script on mythtv system)
icon_dir_local = "icons"
# database settings - note these are over-ridden by ~/.mythtv/mysql.txt (not db_port)
db_host = 'localhost'
db_port = 3306
db_name = 'mythconverg'
db_user = 'mythtv'
db_passwd = 'mythtv'
# # # # # # End of User Settings # # # # # #

# # # # # # Program Settings Here # # # # # #
# # You probably don't want to change these
# How old does the data for a region have to be before we update it (days)
web_cache_period = 1
# How old does an icon have to be before we update it (days)
icon_cache_period = 1
# Folder for region files
pklz_folder = 'regions'
# Courtesy parameter This puts delays in so as not to nark the webmaster (maybe)
# You can set it to 0, but if everyone does that the server might not like us.
courtesy = 1
# # # # # # End of Program Settings # # # # # #

# Libraries required by program
import MySQLdb, difflib, csv
from HTMLParser import HTMLParser, HTMLParseError
import urllib, urlparse
import sys, os.path, time, getopt,  pprint, zlib, cPickle, StringIO, ConfigParser

# Region Superclass. Needs to be first
class Region:
    '''The Region super class. Any from of grabber must override grab_data'''
    def __init__(self, region_code, region_type):
        #initialise structures
        self.region_code = region_code
        self.region_type = region_type
        self.presets = {}
        self.webdata = {}
        self.last_webdata_update = 0
        
    def grab_data(self, force_update):
        '''Grab data from somewhere and structure for use'''
        raise NotImplemented

    def get_presets(self, search_terms):
        '''Search for presets from pklz file that have previously been chosen'''
        results = []
        for idx in range(len(search_terms)):
            term = search_terms[idx]
            if self.presets.has_key(term):
                if self.presets[term] not in [entry['preset_url'] for entry in results]:
                    results.append({
                                'preset_type': self.region_type,
                                'preset_code': self.region_code,
                                'preset_term': term,
                                'preset_url': self.presets[term]
                                })
        return results
        
    def search(self, search_terms, no_of_results):
        '''Performs a search (vaguely fuzzy string matches) against web data'''
        # Raise cutoff if you want to tighten results up
        cutoff = 0.3
        results = []
        seq_matcher = difflib.SequenceMatcher()
        for i in range(len(search_terms)):
            term = search_terms[i]
            seq_matcher.set_seq2(''.join([char for char in term if char not in ' ._']))
            for key in self.webdata.keys():
                seq_matcher.set_seq1(''.join([char for char in key if char not in ' ._']))
                if seq_matcher.real_quick_ratio() >= cutoff \
                            and seq_matcher.quick_ratio() >= cutoff \
                            and seq_matcher.ratio() >= cutoff:
                    results.append([seq_matcher.ratio(), {
                                'pick_ratio': seq_matcher.ratio(), 
                                'pick_key': key,
                                'pick_type': self.region_type,
                                'pick_region': self.region_code,
                                'pick_url': self.webdata[key]
                                }])
        uniq_results = []
        for result in results:
            if result[1]['pick_url'] not in [pick[1]['pick_url'] for pick in uniq_results]:
                uniq_results.append(result)
        results = uniq_results
        results.sort()
        results.reverse()
        return results[:no_of_results]
            
    def dump_data(self):
        '''Aid for debugging'''
        pprint.pprint(self.region_code)
        pprint.pprint(self.presets)
        pprint.pprint(self.webdata)

# # # Start of Subclassed Regions and associated Parsers
# Start of the LyngSat Region class' and parser
class LyngSatCountryRegion(Region):
    '''Region subclass for LyngSat by country/region code'''
    def __init__(self, region_code, region_type):
        #initialise structures
        Region.__init__(self, region_code, region_type)
        self.lyngsat_sub_page = 'http://www.lyngsat-logo.com/tvcountry/%s_%d.html'
        self.lyngsat_type = 'country'
        
    def grab_data(self, force_update):
        '''Grab data from LyngSat and structure for use'''
        # grab data into structure
        high_page = '?'
        cache_expired = time.time() > (self.last_webdata_update + (web_cache_period *24*60*60))
        if force_update or cache_expired:
            # do an update rather than using cached data
            page_no = 1
            while 1:
                sys.stdout.write('Processing web page %d (of %s)...' % (page_no, high_page))
                while 1:
                    try:
                        opener = urllib.FancyURLopener({})
                        page = opener.open(self.lyngsat_sub_page % (self.region_code, page_no))
                        page_contents = page.read()
                        page.close()
                        parser = LyngSatParser(self.webdata, self.region_code, page_no, self.lyngsat_sub_page, self.lyngsat_type)
                        parser.feed(page_contents)
                        break
                    except HTMLParseError, message:
                        print '\n    %s' % (message)
                        sys.stdout.write('    Trying web page %d (of %s) again...' % (page_no, high_page))
                    time.sleep(courtesy)
                sys.stdout.write(' Done.\n')
                time.sleep(courtesy)
                if parser.last_page:
                    break
                page_no = page_no+1
                high_page = parser.high_page
            self.last_webdata_update = time.time()
        self.common_prefix = os.path.dirname(os.path.commonprefix([item[1] for item in self.webdata.items()]))

class LyngSatSatelliteRegion(LyngSatCountryRegion):
    '''Region subclass for LyngSat by satellite/region code'''
    def __init__(self, region_code, region_type):
        #initialise structures
        LyngSatCountryRegion.__init__(self, region_code, region_type)
        self.lyngsat_sub_page = 'http://www.lyngsat-logo.com/tvsat/%s_%d.html'
        self.lyngsat_type = 'sat'

class LyngSatProviderRegion(LyngSatCountryRegion):
    '''Region subclass for LyngSat by satellite/region code'''
    def __init__(self, region_code, region_type):
        #initialise structures
        LyngSatCountryRegion.__init__(self, region_code, region_type)
        self.lyngsat_sub_page = 'http://www.lyngsat-logo.com/tvpack/%s_%d.html'
        self.lyngsat_type = 'pack'

class LyngSatParser(HTMLParser):
    '''Class to parse the retrieved LyngSat pages for likely icons'''
    def __init__(self, results, region_code, this_page, lyngsat_sub_page, lyngsat_type):
        #initialise structures
        HTMLParser.__init__(self)
        self.results = results
        self.last_page = 0
        self.high_page = 1
        self.region_code = region_code
        self.this_page = this_page
        self.lyngsat_sub_page = lyngsat_sub_page
        self.lyngsat_type = lyngsat_type
        
    def handle_starttag(self, tag, attrs):
        '''Called whenever a new tag is encountered (only one we're interested in)'''
        if tag == 'img':
            for pair in attrs:
                if pair[0] == 'src' and pair[1].find('/logo/') != -1:
                    icon_abs_url = urlparse.urljoin(self.lyngsat_sub_page, pair[1])
                    icon_name = os.path.splitext(os.path.basename(icon_abs_url))[0].upper()
                    self.results[icon_name] = icon_abs_url
        # use "<A HREF="../tvcountry/uk_1.html">" to parse the highest page no
        elif tag == 'a':
            for pair in attrs:
                if pair[0] == 'href' and pair[1].find('../tv%s/%s_'%(self.lyngsat_type, self.region_code)) != -1:
                    link_page = int(pair[1][len('../tv%s/%s_'%(self.lyngsat_type, self.region_code)):-5])
                    self.high_page = max(self.high_page, link_page)
                    if self.high_page == self.this_page:
                        self.last_page = 1
# End of the LyngSat Region class' and parser
# Start of the User Region class' and parser
class UserRegion(Region):
    '''Region subclass for user supplied data'''        
    def grab_data(self, force_update):
        '''Dummy method to prevent UserRegion raising NotImplemented'''
        # These UserRegions are for holding single shot imported data as a result of an ImportRegion.
        # As such there is no concept of grabbing the webdata. Just presets, i.e.
        # self.presets = {
        #           <search term 1>: <url 1>,
        #           <search term 2>: <url 2>
        #           }
        pass
# End of the User Region class' and parser
# Start of the Import Region class
class ImportRegion:
    '''Region subclass for importing data from a file'''
    def __init__(self, region, region_list):
        #initialise structures
        self.imported_region = None
        self.import_label = region['import_label']
        import_methods = {
                    'csv': self.parse_csv,
                    'pklz': self.parse_pklz,
                    }
        # Request filename and check for existence
        resp = None
        while 1:
            print 'Please enter import filename for "%s" (<Return> to skip)'  % (self.import_label)
            resp = raw_input('File: ')
            if os.path.exists(resp) and os.path.isfile(resp):
                self.source_name = resp
                break
        # Parse suffix.
        suffix = os.path.splitext(self.source_name)[1][1:]
        # Call appropriate method
        if import_methods.has_key(suffix):
            import_methods[suffix]()
        # Merge or add to main region_list
        if self.imported_region:
            target_region = region_list.get_region(self.imported_region.region_type, self.imported_region.region_code)
            if not target_region:
                target_region = region_classes[self.imported_region.region_type](
                            self.imported_region.region_code,
                            self.imported_region.region_type
                            )
            target_region.presets.update(self.imported_region.presets)
            region_list.region_list.append(target_region)
        print '\nThese imported presets will be saved to %s_%s.pklz' % (
                    self.imported_region.region_type,
                    self.imported_region.region_code,
                    )
        print 'You can use them in future runs by adding "%s:%s" to the regions option.' % (
                    self.imported_region.region_type,
                    self.imported_region.region_code,
                    )

    def parse_pklz(self):
        '''Parse a pklz file'''
        region_file = open(self.source_name, 'r')
        zdata = region_file.read()
        region_file.close()
        data = zlib.decompress(zdata)
        self.imported_region = cPickle.loads(data)

    def parse_csv(self):
        '''Parse a CSV file'''
        csv_presets = {}
        csv_file = open(self.source_name, 'rb')
        reader = csv.reader(csv_file)
        for row in reader:
            if len(row) == 2 and row[1].startswith('http://'):
                csv_presets[row[0].upper()] = row[1]
        csv_file.close()
        self.imported_region = UserRegion(self.import_label, 'user')
        self.imported_region.presets = csv_presets
# End of the Import Region class
# # # Start of Subclassed Regions and associated Parsers

# Register the different Regions available
region_classes = {
            'lsc': LyngSatCountryRegion,
            'lss': LyngSatSatelliteRegion,
            'lsp': LyngSatProviderRegion,
            'user': UserRegion,
            }

# Rest of main code
class RegionList:
    '''Class for managing list of Regions'''
    def __init__(self, regions, overwrite, force_update):
        #initialise structures
        self.overwrite = overwrite
        self.force_update = force_update
        self.region_list = []
        self.regions = self.analyse_regions(regions)
        self.load_regions()
    
    def get_region(self, req_type, req_code):
        '''Return the Region that matches parameters'''
        for entry in [[region.region_type, region.region_code, region] for region in self.region_list]:
            if entry[0] == req_type and entry[1] == req_code:
                return entry[3]
        return None
        
    
    def analyse_regions(self, regions):
        '''Breaks down the passed argument of different classes and regions'''
        results = []
        imports = []
        regions = regions.split(',')
        for region in regions:
            region_type, region_code = region.split(':')
            if region_type != 'import':
                results.append({
                            'region_type': region_type,
                            'region_code': region_code,
                            'region_class': region_classes[region_type]
                            })
            else:
                imports.append({
                            'region_type': region_type,
                            'import_label': region_code,
                            })
        return results+imports
        
    def load_regions(self):
        '''Attempts to load previously saved region files, otherwise creates new ones'''
        for region in self.regions:
            if region['region_type'] != 'import':
                if self.get_region(region['region_type'], region['region_code']):
                    continue
                try:
                    if self.overwrite:
                        # create blank region to overwrite old
                        print 'Ignoring data for region "%s", type "%s"...' % (region['region_code'], region['region_type'])
                        self.region_list.append(region['region_class'](region['region_code'], region['region_type']))
                    else:
                        # load region
                        self.region_list.append(self.load_region(region['region_code'], region['region_type']))
                        print 'Using data for region "%s", type "%s"...' % (region['region_code'], region['region_type'])
                except IOError:
                    # otherwise we create a new region
                    print 'No previous data for region "%s", type "%s", creating new data...' % (
                                region['region_code'], region['region_type'])
                    self.region_list.append(region['region_class'](region['region_code'], region['region_type']))
            else:
                ImportRegion(region, self)
        for region in self.region_list:
            region.grab_data(self.force_update)
    
    def load_region(self, region_code, region_type):
        '''Open file, unzip and unpickle it, and append to self.region_list'''
        region_file_name = os.path.join(pklz_folder, '%s_%s.pklz' % (region_type, region_code))
        region_file = open(region_file_name, 'r')
        zdata = region_file.read()
        region_file.close()
        data = zlib.decompress(zdata)
        return cPickle.loads(data)
                
    def save_regions(self):
        '''Save each region in turn'''
        if os.path.exists(pklz_folder) and not os.path.isdir(pklz_folder):
            print 'Erm... pklz_folder (%s) needs to be a folder...'
            sys.exit(2)
        if not os.path.exists(pklz_folder):
            os.mkdir(pklz_folder)
        if self.overwrite:
            # warn user and ask confirmation before destroying any previous files
            print 'Warning: You have selected to overwrite previous region files. Confirming will'
            print 'lose previous details in region files. Anything else will lose all updates in'
            print 'this running of the script.'
            resp = raw_input('Do you wish to overwrite old region files [y/Y]? ')
            if resp.upper() != 'Y':
                print 'Not overwriting previous region files. Changes will be lost.'
                return
        for region in self.region_list:
            self.save_region(region)
            
    def save_region(self, region):
        '''Pickle, gzip and save region to a file'''
        region_file_name = os.path.join(pklz_folder, '%s_%s.pklz' % (region.region_type, region.region_code))
        data = cPickle.dumps(region)
        zdata = zlib.compress(data)
        print 'Updating/replacing "%s" region file (%s,%d,%d,%d%%)' % (
                    region.region_code, region_file_name, len(data),
                    len(zdata), int((100*len(zdata))/len(data))
                    )
        region_file = open(region_file_name, 'w')
        region_file.write(zdata)
        region_file.close()
        
    def get_presets(self, search_terms):
        '''Return a unified list of presets'''
        results = []
        for region in self.region_list:
            results = results+region.get_presets(search_terms)
        return results
        
    def search_regions(self, search_terms, no_of_results=10):
        '''Search each Region in turn for search terms'''
        results = []
        single_exact_match = 0
        for region in self.region_list:
            results = results+region.search(search_terms, no_of_results)
        uniq_results = []
        for result in results:
            if result[1]['pick_url'] not in [pick[1]['pick_url'] for pick in uniq_results]:
                uniq_results.append(result)
        results = uniq_results
        results.sort()
        results.reverse()
        results = results[:no_of_results]
        results = [entry[1] for entry in results]
        if results and results[0]['pick_ratio'] == 1 and results[1]['pick_ratio'] != 1:
            single_exact_match = 1
        return (results, single_exact_match)

    def update_region(self, channel_set, searchable_term):
        '''Stores the preset information prior to writing to file'''
        for region in self.region_list:
            if region.region_type == channel_set['pick_type'] and region.region_code == channel_set['pick_region']:
                region.presets[searchable_term] = channel_set['pick_url']
    
    def dump_data(self):
        '''Aid for debugging'''
        for region in self.region_list:
            region.dump_data()
        

class ChannelList:
    '''Class for managing list of Channels'''
    def __init__(self, sourceid, force_update):
        #initialise structures
        self.force_update = force_update
        self.region_list = None
        dbase = MySQLdb.connect(host=db_host, port=db_port, db=db_name, user=db_user, passwd=db_passwd)
        cursor = dbase.cursor()
        query = '''SELECT chanid, name, callsign, xmltvid, icon FROM channel'''
        if sourceid:
            query = query + ''' WHERE sourceid = "%s"''' % (sourceid)
# Not done here now, but ggod to keep in case another idea for limited things comes along.            
##        if unassigned:
##            if sourceid:
##                query = query + ''' AND'''
##            query = query + ''' WHERE icon = NULL''' % (sourceid)
        query = query + ''';'''
        cursor.execute(query)
        inter_results = cursor.fetchall()
        cursor.close()
        self.results = []
        for entry in inter_results:
            self.results.append({
                        'db_id': entry[0],
                        'db_name': entry[1],
                        'db_callsign': entry[2],
                        'db_xmltvid': entry[3],
                        'db_icon': entry[4]
                        })
            
    def dump_data(self):
        '''Aid for debugging'''
        pprint.pprint(self.results)
        
    def retrieve_icons(self):
        '''Retrieve icons from the web'''
        local_results = []
        for result in self.results:
            if result.has_key('pick_url'):
                local_results.append(result)
        channel_count = len(local_results)
        channel_current = 0
        for channel_set in local_results:
            channel_current = channel_current+1
            local_folder = os.path.join(icon_dir_local, channel_set['pick_type'], channel_set['pick_region'])
            if not os.path.exists(local_folder):
                os.makedirs(local_folder)
            local_file = os.path.join(local_folder, os.path.basename(channel_set['pick_url']))
            if os.path.exists(local_file):
                local_file_timestamp = os.stat(local_file)[8]
            else:
                local_file_timestamp = 0
            local_file_expired = time.time() > (local_file_timestamp + (icon_cache_period *24*60*60))
            
            remote_file = channel_set['pick_url']
            if remote_file is not None \
                        and remote_file.startswith('http://') \
                        and (local_file_expired or self.force_update):
                print 'Retrieving icon %d of %d...' % (channel_current, channel_count)
                urllib.urlretrieve(remote_file, local_file)
            else:
                print 'Not retrieving icon %d of %d. Either not changed or already cached.' % (channel_current, channel_count)
            time.sleep(courtesy)
    
    def update_dbase_callsign(self, dtf):
        '''Update the DB with, or write to file, information to allow mythfilldatabase to retrieve icons'''
        # Note that this is very experimental. I have no way to properly test this. Assistance would be appreciated.
        if not dtf:
            dbase = MySQLdb.connect(host=db_host, port=db_port, db=db_name, user=db_user, passwd=db_passwd)
            cursor = dbase.cursor()
        else:
            cursor = None
        query = '''DELETE FROM settings WHERE value like 'mythfilldatabase.urlmap.%';\n'''
        self.write_dbase_query(dtf, cursor, query)
        query = '''DELETE FROM callsignnetworkmap;\n'''
        self.write_dbase_query(dtf, cursor, query)
        query = '''DELETE FROM networkiconmap;\n'''
        self.write_dbase_query(dtf, cursor, query)
        
        for channel_set in self.results:
            if channel_set.has_key('pick_url') and channel_set['pick_url'].startswith("http://"):
                region = self.region_list.get_region(channel_set['pick_type'], channel_set['pick_region'])
                db_urlmap = '[%s_%s]' % (region.region_type, region.region_code)
                query = '''INSERT INTO settings VALUES ('mythfilldatabase.urlmap.%s','%s',NULL);\n''' % (
                            db_urlmap,
                            region.common_prefix
                            )
                self.write_dbase_query(dtf, cursor, query)
                if channel_set['pick_url'].startswith(region.common_prefix):
                    url = db_urlmap+channel_set['pick_url'][len(region.common_prefix):]
                else:
                    url = channel_set['pick_url']
                chan_callsign = channel_set['db_callsign']
                query = '''INSERT INTO callsignnetworkmap (callsign,network) VALUES ('%s','%s_%s');\n''' % (
                            chan_callsign,
                            db_urlmap,
                            chan_callsign
                            )
                self.write_dbase_query(dtf, cursor, query)
                query = '''INSERT INTO networkiconmap (network,url) VALUES ('%s_%s','%s');\n''' % (
                            db_urlmap,
                            chan_callsign,
                            url
                            )
                self.write_dbase_query(dtf, cursor, query)
        if cursor:
            cursor.close()
    
    def update_dbase_channel(self, sourceid, dtf):
        '''Update the DB with, or write to file, information on locations of icons as determined by this script'''
        if not dtf:
            dbase = MySQLdb.connect(host=db_host, port=db_port, db=db_name, user=db_user, passwd=db_passwd)
            cursor = dbase.cursor()
        else:
            cursor = None
        for channel_set in self.results:
            if channel_set['db_icon'] == None:
                query = '''UPDATE channel SET icon=NULL WHERE chanid="%s"''' % (channel_set['db_id'])
                if sourceid is not None:
                    query = query + ''' AND sourceid=%d;\n''' % (sourceid)
                else:
                    query = query + ''';\n'''
                self.write_dbase_query(dtf, cursor, query)
            elif channel_set.has_key('pick_url') and channel_set['pick_url'].startswith("http://"):
                new_path = os.path.join(
                            icon_dir_mythtv, channel_set['pick_type'],
                            channel_set['pick_region'],
                            os.path.basename(channel_set['pick_url'])
                            )
                query = '''UPDATE channel SET icon="%s" WHERE chanid="%s"''' % (new_path, channel_set['db_id'])
                if sourceid is not None:
                    query = query + ''' AND sourceid=%d;\n''' % (sourceid)
                else:
                    query = query + ''';\n'''
                self.write_dbase_query(dtf, cursor, query)
        if cursor:
            cursor.close()
        
    def write_dbase_query(self, dtf, cursor, query):
        '''Either send query to DB, or write to file'''
        if dtf:
            dtf.write(query)
        else:
            cursor.execute(query)

    def update_regions(self, region_list, callsigns):
        '''Update appropriate Region for each channel'''
        for channel_set in self.results:
            if channel_set.has_key('pick_url') and channel_set['pick_url'].startswith("http://"):
                if callsigns:
                    searchable_term = channel_set['db_callsign']     # Store URL against callsign
                elif channel_set.has_key('db_xmltvid') and channel_set['db_xmltvid']:
                    searchable_term = channel_set['db_xmltvid']     # Store URL against xmltvid if exists
                else:
                    searchable_term = channel_set['db_name']     # Store URL against channel name
                region_list.update_region(channel_set, searchable_term.upper())
    
    def print_info(self):
        '''Print a list of all Channels as determined by options, with their current values'''
        count = len(self.results)
        for idx in range(count):
            print '[%03d/%03d] %s: %-20s %-20s %s' % (
                        idx+1, count,
                        self.results[idx]['db_id'],
                        self.results[idx]['db_name'],
                        self.results[idx]['db_callsign'],
                        self.results[idx]['db_xmltvid']
                        )
    
class Matcher:
    '''The main icon matcher class, run on creation'''
    def __init__(self, channel_list, region_list, no_auto_select, unassigned, limit):
        #initialise structures
        self.channel_list = channel_list
        self.region_list = region_list
        self.no_auto_select = no_auto_select
        self.unassigned = unassigned
        self.limit = limit
        if self.limit:
            self.limit.sort()
        self.assign_icons()
        
    def assign_icons(self):
        '''The main control method for selecting icons per channel'''
        exit_flag = 0
        autospeed = 0
        channel_targets = self.filter_channels()
        channel_index = -1
        for channel_set in self.channel_list.results:
            channel_index = channel_index+1
            if channel_index not in channel_targets:
                continue
            search_terms = [term.upper() for term in [
                        channel_set['db_name'],
                        channel_set['db_callsign'],
                        channel_set['db_xmltvid']
                        ]]
            while exit_flag == 0:
                resp = None
                self.display_header(channel_set, channel_index, channel_targets)
                # Find exact matches in previous entries
                presets = self.region_list.get_presets(search_terms)
                if not len(presets):
                    print 'I have no matching presets.'
                elif len(presets) == 1:
                    print 'Found one exact matching preset for %s, in %s:%s' % (
                                presets[0]['preset_term'],
                                presets[0]['preset_type'],
                                presets[0]['preset_code']
                                )
                else:
                    print 'Found multiple matching presets. Selecting will give menu.'
                
                # Find fuzzy matches
                (option_list, single_exact_match)=self.region_list.search_regions(search_terms)
                if len(option_list) > 0:
                    print 'The best fuzzy matches (up to ten) are:'
                    num_prompt = '1'
                    if len(option_list) > 1:
                        num_prompt = num_prompt+'-%d' % len(option_list)
                    num_prompt = num_prompt+'/'
                    if single_exact_match and not self.no_auto_select:
                        resp = '1'
                        print 'Auto Selecting exact match (entry 1)'
                else:
                    print 'No suitable fuzzy matches found!'
                    num_prompt = ''
                
                # build fuzzy matches menu
                for idx in range(0, len(option_list)):
                    url_tail = urlparse.urlunparse(tuple(['', '']+list(urlparse.urlparse(option_list[idx]['pick_url'])[2:])))
                    entry = '  %2d. %s:%s, %-60s' % (
                                idx+1,
                                option_list[idx]['pick_type'],
                                option_list[idx]['pick_region'],
                                option_list[idx]['pick_key']
                                )
                    entry = (entry[:40] + '  (%s)%s' % (url_tail, ' '*40))[:80]
                    if entry[79] not in ') ':
                        entry = entry[:77] + '...'
                    print entry
                
                # get user response
                num_resp_options = []
                for entry in range(1, len(option_list)+1):
                    num_resp_options.append(str(entry))
                resp_options = len(presets) and ['p'] or []
                resp_options = resp_options + ['f', 'u', 'a', 'c', '.', ''] + num_resp_options
                
                # Where we have a single preset move through quickly
                if autospeed:
                    if channel_set['db_icon'] == '':
                        if len(presets):
                            resp = 'p'
                        else:
                            autospeed = 0
                while resp == None:
                    # get interactive user response
                    print """Enter number to pick icon from list, "f" for new fuzzy term, "u" to enter URL,"""
                    print """"a" to %s autospeed, "c" to clear, %s""" % (
                                autospeed and 'stop' or 'start',
                                len(presets) and '"p" to select preset list entry,' or ''
                                )
                    print """"." to skip all remaining channels or <Return> to skip this channel."""
                    resp = raw_input('Please enter choice [%sf/u/a/c/%s./<Return>]? ' % (num_prompt, len(presets) and 'p/' or ''))
                    if resp not in resp_options:
                        resp = None
                        print
                if resp in num_resp_options:
                    # Selection of one of the fuzzy matches
                    channel_set.update(option_list[int(resp)-1])
                    break
                if resp == 'u':
                    # Enter a manual URL
                    print 'Note that there is no checking done on the entered URL.'
                    channel_set['pick_url'] = raw_input('Please enter URL: ')
                    break
                if resp == 'a':
                    # Activate autospeed
                    if autospeed:
                        autospeed = 0
                    else:
                        autospeed = 1
                if resp == 'p':
                    # Select a preset
                    if len(presets) == 1:
                        channel_set['pick_url'] = presets[0]
                        break
                    else:
                        # get choice from user
                        print '\n\n[%03d/%03d, Index: %03d] Channel id: %s is currently set to:' % (
                                    channel_targets.index(channel_index)+1,
                                    len(channel_targets),
                                    channel_index+1,
                                    channel_set['db_id']
                                    )
                        print '        Name: %s' % (channel_set['db_name'])
                        print '    Callsign: %s' % (channel_set['db_callsign'])
                        print '    XMLTV ID: %s' % (channel_set['db_xmltvid'])
                        print '   Icon Path: %s' % (channel_set['db_icon'])
                        print
                        print 'The available presets are:'
                        for idx in range(0, len(presets)):
                            url_tail = urlparse.urlunparse(tuple(['', '']+list(urlparse.urlparse(presets[idx]['preset_url'])[2:])))
                            entry = '  %2d. %s:%s, %-60s' % (
                                        idx+1,
                                        presets[idx]['preset_type'],
                                        presets[idx]['preset_code'],
                                        presets[idx]['preset_term'])
                            entry = (entry[:40] + '  (%s)%s' % (url_tail, ' '*40))[:80]
                            if entry[79] not in ') ':
                                entry = entry[:77] + '...'
                            print entry
                        num_resp_options = []
                        for entry in range(1, len(presets)+1):
                            num_resp_options.append(str(entry))
                        resp_options = ['r']+num_resp_options
                        resp = None
                        while resp == None:
                            print """Enter a number to select icon from list, or "r" to return to main menu."""
                            resp = raw_input('Please enter choice [1-%d/r]? ' % (len(presets)))
                            if resp not in resp_options:
                                resp = None
                                print
                            if resp in num_resp_options:
                                channel_set['pick_url'] = presets[int(resp)-1]['preset_url']
                                ret_main = 0
                            if resp == 'r':
                                ret_main = 1
                        if not ret_main:
                            break
                if resp == 'c':
                    # Clear this channel
                    channel_set['db_icon'] = None
                    break
                if resp == 'f':
                    # Enter a new fuzzy term
                    search_term = raw_input('Enter a new fuzzy term: ')
                    search_terms = [search_term.upper()]
                if resp == '':
                    # Skip this channel
                    break
                if resp == '.':
                    # Exit out of entering any further channels
                    print
                    exit_flag = 1
                    break

    def filter_channels(self):
        '''Run through various filters, pairing the list of Channels down'''
        channel_targets = range(len(self.channel_list.results))
        for target in channel_targets[:]:
            if self.limit and (target+1) not in self.limit:
                channel_targets.remove(target)
            if self.unassigned and self.channel_list.results[target]['db_icon'] != '' and (target+1) in channel_targets:
                channel_targets.remove(target)
        return channel_targets

    def display_header(self, channel_set, channel_index, channel_targets):
        '''Write the main banner header display'''
        headerstring = 80*'-'
        headerstring_offset = 2
        headerstring_left = 'Channel ID: %s' % (channel_set['db_id'])
        headerstring_right = '[%03d/%03d, Index: %03d]' % (
                    channel_targets.index(channel_index)+1,
                    len(channel_targets),
                    channel_index+1
                    )
        headerstring = (
                    headerstring[0:headerstring_offset]+
                    headerstring_left+
                    headerstring[len(headerstring_left)+headerstring_offset:-(len(headerstring_right)+headerstring_offset)]+
                    headerstring_right+
                    headerstring[-headerstring_offset:]
                    )
        print '\n\n'+headerstring
        print '        Name: %s' % (channel_set['db_name'])
        print '    Callsign: %s' % (channel_set['db_callsign'])
        print '    XMLTV ID: %s' % (channel_set['db_xmltvid'])
        print '   Icon Path: %s' % (channel_set['db_icon'])
        print
        
def usage():
    '''Small stub to display the module docstring (see top of file)'''
    print __doc__ % __revision__

def erase_dbase_icons():
    '''Quick function to remove icon setting for all channels in DB'''
    dbase = MySQLdb.connect(host=db_host, port=db_port, db=db_name, user=db_user, passwd=db_passwd)
    cursor = dbase.cursor()
    query = '''UPDATE channel SET icon=NULL;'''
    cursor.execute(query)
    cursor.close()
    
def main():
    '''Main program control'''
    # Section to grab access settings from mysql.txt if it exists
    myth_db_config_file = os.path.expanduser('~/.mythtv/mysql.txt')
    if os.path.exists(myth_db_config_file) and os.path.isfile(myth_db_config_file):
        conf_parser = ConfigParser.SafeConfigParser()
        conf_parser.readfp(StringIO.StringIO('[MythDBSettings]\n'+open(myth_db_config_file).read()))
        myth_db_config = dict(conf_parser.items('MythDBSettings'))
        global db_host, db_name, db_user, db_passwd
        db_host = myth_db_config.get('dbhostname', db_host)
        db_name = myth_db_config.get('dbname', db_name)
        db_user = myth_db_config.get('dbusername', db_user)
        db_passwd = myth_db_config.get('dbpassword', db_passwd)
        
    # extract arguments
    try:
        opts, _ = getopt.getopt(
                    sys.argv[1:],
                    'hr:s:ocfnd:il:eu',
                    ['help', 'regions=', 'source=', 'overwrite', 'callsigns', 'force-update', 'no-auto-select', 'dump', 'info',
                      'limit=', 'erase', 'unassigned']
                    )
    except getopt.GetoptError:
        # print help information and exit:
        print 'GetOpt Error'
        usage()
        sys.exit(2)
    
    # Default settings
    global default_region_codes
    region_codes = default_region_codes
    sourceid = None
    overwrite = 0
    callsigns = 0
    force_update = 0
    no_auto_select = 0
    dtf_name = None
    info = 0
    limit = None
    erase = 0
    unassigned = 0
    
    # Process cmd line options
    for opt, arg in opts:
        if opt in ('-h', '--help'):
            usage()
            sys.exit()
        if opt in ('-r', '--regions'):
            region_codes = arg
        if opt in ('-s', '--source'):
            sourceid = arg
        if opt in ('-o', '--overwrite'):
            overwrite = 1
        if opt in ('-c', '--callsigns'):
            callsigns = 1
        if opt in ('-f', '--force-update'):
            force_update = 1
        if opt in ('-n', '--no-auto-select'):
            no_auto_select = 1
        if opt in ('-d', '--dump'):
            dtf_name = os.path.abspath(arg)
            if not os.path.exists(os.path.dirname(dtf_name)):
                print 'Can\'t find folder for SQL dump file. Please check path.'
                raise
        if opt in ('-i', '--info'):
            info = 1
        if opt in ('-l', '--limit'):
            limit = []
            prelimit = arg
            prelimit = prelimit.split(',')
            for item in prelimit:
                if '-' in item:
                    pair = item.split('-')
                    if len(pair) > 2:
                        raise Exception('Too many parts in a range (x-y).')
                    rng = range(int(pair[0]), int(pair[1])+1)
                    limit = limit+rng
                else:
                    limit.append(int(item))
        if opt in ('-e', '--erase'):
            erase = 1
        if opt in ('-u', '--unassigned'):
            unassigned = 1
    
    # Erase database icon settings
    if not info and erase:
        erase_dbase_icons()
    
    # Get DB channels
    channel_list = ChannelList(sourceid, force_update)
    #channel_list.dumpData()
    
    if info:
        channel_list.print_info()
        sys.exit(0)
    
    # Open the previous data files
    region_list = RegionList(region_codes, overwrite, force_update)
    channel_list.region_list = region_list
    #region_list.dumpData()    
    
    # Assign icons
    Matcher(channel_list, region_list, no_auto_select, unassigned, limit)
    #channel_list.dumpData()
    
    # The useful bit :-)
    if dtf_name:
        dtf = open(dtf_name, 'w')
    else:
        dtf = None
    if callsigns:
        # Update DB callsignnetworkmap, networkiconmap, settings
        channel_list.update_dbase_callsign(dtf)
    else:
        # Retrieve icons
        channel_list.retrieve_icons()
        # Update DB channel
        channel_list.update_dbase_channel(sourceid, dtf)
    if dtf:
        dtf.close()
    
    #region_list.dumpData()
    
    # Update data held for the regions
    channel_list.update_regions(region_list, callsigns)
    
    # Write our updates to the local data files
    region_list.save_regions()
    
if __name__ == '__main__':
    main()
    sys.exit(0)
